import logging
from typing import ClassVar

import click


class ColorFormatter(logging.Formatter):
    colors: ClassVar[dict] = {
        "DEBUG": {"fg": "blue"},
        "WARNING": {"fg": "yellow"},
        "ERROR": {"fg": "red"},
        "CRITICAL": {"fg": "red"},
    }

    def formatMessage(self, record):  # noqa: N802
        level = record.levelname.upper()

        msg = record.getMessage()

        indentation = getattr(record, "indentation", None)
        msg = "\n".join("  " * (indentation or 0) + x for x in msg.splitlines())

        if level in self.colors:
            prefix = click.style(f"{level}: ", **self.colors[level])
            msg = "\n".join(prefix + x for x in msg.splitlines())

        if indentation == 0:
            msg = f"\n{msg}"

        return msg


class ClickHandler(logging.Handler):
    def emit(self, record):
        try:
            msg = self.format(record)
            is_error = record.levelname.upper() in [
                "WARNING",
                "ERROR",
                "CRITICAL",
            ]
            click.echo(msg, err=is_error)
        except Exception:  # noqa: no cover
            self.handleError(record)


_default_handler = ClickHandler()
_default_handler.formatter = ColorFormatter()


def configure_logger(logger):
    logger.handlers = [_default_handler]
    logger.propagate = False


def add_verbose_option(loggers):
    """
    A decorator that adds a `--verbose, -v` option to the decorated command and
    configure adds a handler to the logger.
    """

    for logger in loggers:
        configure_logger(logger)

    def decorator(f):
        def set_level(ctx, param, value):  # noqa:ARG001
            for logger in loggers:
                if value:
                    logger.setLevel(logging.DEBUG)
                else:
                    logger.setLevel(logging.INFO)

            return value

        return click.option(
            "--verbose",
            "-v",
            callback=set_level,
            is_flag=True,
            default=False,
            help="Enables verbose mode.",
        )(f)

    return decorator



import enum
from typing import List, Union

from .tables import metadata


class TableSetEnum(enum.Enum):
    """
    Options to control which tables to keep in the database after the import.
    """

    UNIFIED_CEP_ONLY = "unified-cep-only"
    CEP_TABLES = "cep-tables"
    ALL_TABLES = "all"

    @property
    def to_populate(self) -> List[str]:
        if self.value == "all":
            return [t.name for t in metadata.sorted_tables]

        return get_cep_tables()

    @property
    def to_drop(self) -> List[str]:
        if self.value == "unified-cep-only":
            return [
                t.name
                for t in metadata.sorted_tables
                if not t.info.get("unified_table")
            ]

        return []


def get_cep_tables() -> List[str]:
    """
    Get the list of tables that are required for CEP search.
    """
    tables = []
    dependencies = set()

    for table in reversed(metadata.sorted_tables):
        if "cep" in table.c or table.name in dependencies:
            for dep in table.foreign_keys:
                dependencies.add(dep.column.table.name)

            tables.append(table.name)

    return list(reversed(tables))


def get_table_files_glob(table_name: str) -> Union[str, None]:
    """
    Get the file globs for each table.
    Calculate from table name when not specified.
    """
    table = metadata.tables[table_name]

    if table.info.get("unified_table", False):
        return None

    return table.info.get("file_glob", f"{table.name.upper()}.TXT")



import logging
from typing import Iterable, Iterator

import sqlalchemy as sa

from .tables import (
    cep_unificado,
    log_bairro,
    log_cpc,
    log_grande_usuario,
    log_localidade,
    log_logradouro,
    log_unid_oper,
)

logger = logging.getLogger(__name__)


def cep_unificado_insert_from(rows: "sa.Select"):
    return cep_unificado.insert().from_select(rows.selected_columns, rows.subquery())


def normalize_logradouro(rows: "sa.CursorResult") -> Iterator[dict]:
    for row in rows:
        logradouro_parts = row.logradouro.split(",", 1)

        yield {
            **row._asdict(),
            "logradouro": logradouro_parts[0].strip(),
            "complemento": logradouro_parts[1].strip()
            if len(logradouro_parts) > 1
            else None,
        }


def cep_unificado_insert_in_batches(
    conn: sa.Connection, rows: Iterable[dict], batch_size: int = 500
):
    """
    Insert rows in the unified table in batches
    """
    batch = []

    for row in rows:
        batch.append(row)

        if len(batch) == batch_size:
            logger.debug("Inserting %d rows into cep_unificado", len(batch))
            conn.execute(cep_unificado.insert().values(batch))
            batch = []

    if batch:
        logger.debug("Inserting %d rows into cep_unificado", len(batch))
        conn.execute(cep_unificado.insert().values(batch))


def select_logradouros_ceps() -> "sa.Select":
    return sa.select(
        log_logradouro.c.cep.label("cep"),
        sa.case(
            (
                log_logradouro.c.log_sta_tlo == "S",
                log_logradouro.c.tlo_tx + " " + log_logradouro.c.log_no,
            ),
            else_=log_logradouro.c.log_no,
        ).label("logradouro"),
        log_bairro.c.bai_no.label("bairro"),
        log_localidade.c.loc_no.label("municipio"),
        log_localidade.c.mun_nu.label("municipio_cod_ibge"),
        log_logradouro.c.ufe_sg.label("uf"),
    ).select_from(
        log_logradouro.join(log_localidade).join(
            log_bairro, onclause=log_logradouro.c.bai_nu_ini == log_bairro.c.bai_nu
        )
    )


def select_localidades_ceps() -> "sa.Select":
    return (
        sa.select(
            log_localidade.c.cep.label("cep"),
            log_localidade.c.ufe_sg.label("uf"),
            log_localidade.c.loc_no.label("municipio"),
            log_localidade.c.mun_nu.label("municipio_cod_ibge"),
        )
        .select_from(log_localidade)
        .where(
            log_localidade.c.cep.isnot(None) & log_localidade.c.loc_nu_sub.is_(None),
        )
    )


def select_localidades_subordinadas_ceps() -> "sa.Select":
    localidade_subordinada = log_localidade.alias()

    return (
        sa.select(
            log_localidade.c.cep.label("cep"),
            log_localidade.c.ufe_sg.label("uf"),
            localidade_subordinada.c.loc_no.label("municipio"),
            localidade_subordinada.c.mun_nu.label("municipio_cod_ibge"),
            log_localidade.c.loc_no.label("bairro"),
        )
        .select_from(log_localidade)
        .join(
            localidade_subordinada,
            onclause=log_localidade.c.loc_nu_sub == localidade_subordinada.c.loc_nu,
        )
        .where(
            log_localidade.c.cep.isnot(None) & log_localidade.c.loc_nu_sub.isnot(None),
        )
    )


def select_cpc_ceps() -> "sa.Select":
    localidade_subordinada = log_localidade.alias()

    return (
        sa.select(
            log_cpc.c.cep.label("cep"),
            log_cpc.c.cpc_endereco.label("logradouro"),
            sa.func.coalesce(
                localidade_subordinada.c.loc_no,
                log_localidade.c.loc_no,
            ).label("municipio"),
            sa.func.coalesce(
                localidade_subordinada.c.mun_nu,
                log_localidade.c.mun_nu,
            ).label("municipio_cod_ibge"),
            log_cpc.c.ufe_sg.label("uf"),
            log_cpc.c.cpc_no.label("nome"),
        )
        .select_from(log_cpc)
        .join(log_localidade)
        .outerjoin(
            localidade_subordinada,
            onclause=log_localidade.c.loc_nu_sub == localidade_subordinada.c.loc_nu,
        )
    )


def select_grandes_usuarios_ceps() -> "sa.Select":
    localidade_subordinada = log_localidade.alias()

    return (
        sa.select(
            log_grande_usuario.c.cep.label("cep"),
            log_grande_usuario.c.gru_endereco.label("logradouro"),
            log_bairro.c.bai_no.label("bairro"),
            sa.func.coalesce(
                localidade_subordinada.c.loc_no,
                log_localidade.c.loc_no,
            ).label("municipio"),
            sa.func.coalesce(
                localidade_subordinada.c.mun_nu,
                log_localidade.c.mun_nu,
            ).label("municipio_cod_ibge"),
            log_grande_usuario.c.ufe_sg.label("uf"),
            log_grande_usuario.c.gru_no.label("nome"),
        )
        .select_from(log_grande_usuario)
        .join(log_localidade)
        .join(log_bairro, onclause=log_grande_usuario.c.bai_nu == log_bairro.c.bai_nu)
        .outerjoin(
            localidade_subordinada,
            onclause=log_localidade.c.loc_nu_sub == localidade_subordinada.c.loc_nu,
        )
    )


def select_unidades_operacionais_ceps() -> "sa.Select":
    localidade_subordinada = log_localidade.alias()

    return (
        sa.select(
            log_unid_oper.c.cep.label("cep"),
            log_unid_oper.c.uop_endereco.label("logradouro"),
            log_bairro.c.bai_no.label("bairro"),
            sa.func.coalesce(
                localidade_subordinada.c.loc_no,
                log_localidade.c.loc_no,
            ).label("municipio"),
            sa.func.coalesce(
                localidade_subordinada.c.mun_nu,
                log_localidade.c.mun_nu,
            ).label("municipio_cod_ibge"),
            log_unid_oper.c.ufe_sg.label("uf"),
            log_unid_oper.c.uop_no.label("nome"),
        )
        .select_from(log_unid_oper)
        .join(log_localidade)
        .join(log_bairro, onclause=log_unid_oper.c.bai_nu == log_bairro.c.bai_nu)
        .outerjoin(
            localidade_subordinada,
            onclause=log_localidade.c.loc_nu_sub == localidade_subordinada.c.loc_nu,
        )
    )


def populate_unified_table(conn: sa.Connection, insert_batch_size: int = 500):
    """
    Query unifying rows from all tables with CEP address information
    """
    insert_from_selects = [
        (select_logradouros_ceps(), "logradouros"),
        (select_localidades_ceps(), "localidades"),
        (select_localidades_subordinadas_ceps(), "localidades subordinadas"),
    ]

    for select_stmt, name in insert_from_selects:
        logger.info(
            "Populating unified CEP table with %s data",
            name,
            extra={"indentation": 1},
        )
        conn.execute(cep_unificado_insert_from(select_stmt))

        inserted = conn.execute(
            sa.select(sa.func.count()).select_from(select_stmt.subquery())
        ).scalar()

        logger.info(
            "Inserted %s CEPs from %s into table %s",
            inserted,
            name,
            cep_unificado.name,
            extra={"indentation": 2},
        )

    # the following ones need some normalization before inserting
    # they are normalized via python, so this can run for different DBMSs

    selects_with_normalization = [
        (select_cpc_ceps(), "CPC"),
        (select_grandes_usuarios_ceps(), "grandes usuários"),
        (select_unidades_operacionais_ceps(), "unidades operacionais"),
    ]

    for select_stmt, name in selects_with_normalization:
        logger.info(
            "Populating unified CEP table with normalized %s data",
            name,
            extra={"indentation": 1},
        )

        cep_unificado_insert_in_batches(
            conn,
            normalize_logradouro(conn.execute(select_stmt).yield_per(1000)),
            batch_size=insert_batch_size,
        )

        inserted = conn.execute(
            sa.select(sa.func.count()).select_from(select_stmt.subquery())
        ).scalar()

        logger.info(
            "Inserted %s CEPs from %s into table %s",
            inserted,
            name,
            cep_unificado.name,
            extra={"indentation": 2},
        )

    inserted = conn.execute(
        sa.select(sa.func.count()).select_from(cep_unificado)
    ).scalar()

    logger.info(
        'Inserted %s rows into table "%s"',
        inserted,
        cep_unificado.name,
        extra={"indentation": 1},
    )



import json
import logging
import sys
from typing import Optional

import click

from edne_correios_loader.__about__ import __version__
from edne_correios_loader.cep_querier import CepQuerier
from edne_correios_loader.dbwriter import logger as dbwriter_logger
from edne_correios_loader.loader import DneLoader
from edne_correios_loader.loader import logger as loader_logger
from edne_correios_loader.resolver import DneResolver
from edne_correios_loader.resolver import logger as resolver_logger
from edne_correios_loader.table_set import TableSetEnum
from edne_correios_loader.unified_table import logger as unified_table_logger

from .logger import add_verbose_option

logger = logging.getLogger(__name__)


class DneResolverWithDownloadProgress(DneResolver):
    progress_bar: Optional[click.progressbar] = None

    def download_report_hook(self, read, total, hook_type):
        if total == -1:
            return

        if hook_type == "start":
            self.progress_bar = click.progressbar(
                length=total, label="Downloading DNE file"
            )
            self.progress_bar.render_progress()

        if read:
            self.progress_bar.update(read)

        if hook_type == "finish":
            self.progress_bar.finish()
            self.progress_bar.render_finish()


class DneLoaderWithProgress(DneLoader):
    DneResolver = DneResolverWithDownloadProgress


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(version=__version__, prog_name="DNE Correios Loader")
def edne_correios_loader():
    ...  # pragma: no cover


@edne_correios_loader.command()
@click.option(
    "-s",
    "--dne-source",
    help="Path or URL with the DNE file/dir to be imported",
    metavar="<path/zip-file/url>",
)
@click.option(
    "-db",
    "--database-url",
    help="Database URL where the DNE data will be imported to",
    required=True,
    metavar="<url>",
)
@click.option(
    "--tables",
    type=click.Choice(
        [option.value for option in list(TableSetEnum)],
        case_sensitive=False,
    ),
    help="Which tables to keep in the database after the import",
    default="unified-cep-only",
)
@add_verbose_option(
    [logger, loader_logger, resolver_logger, dbwriter_logger, unified_table_logger]
)
def load(dne_source, database_url, tables, verbose):
    """
    Load DNE data into a database.
    """

    click.echo(click.style(f"Starting DNE Correios Loader v{__version__}", bold=True))

    try:
        DneLoaderWithProgress(database_url, dne_source=dne_source).load(
            table_set=TableSetEnum(tables)
        )
    except Exception as e:
        if verbose:
            logger.exception(e)  # noqa: TRY401
        else:
            logger.error(e)  # noqa: TRY400

        sys.exit(1)


@click.option(
    "-db",
    "--database-url",
    help="Database URL where the DNE data will be imported to",
    required=True,
    metavar="<url>",
)
@click.argument("cep")
@edne_correios_loader.command()
def query_cep(database_url, cep):
    """
    Query a CEP from the database to ensure it was correctly populated.
    """
    try:
        cep_address = CepQuerier(database_url).query(cep)
    except Exception as e:
        logger.error(e)  # noqa: TRY400
        sys.exit(1)

    if cep_address:
        click.echo(
            click.style(
                json.dumps(cep_address, ensure_ascii=False, indent=2), fg="green"
            )
        )
    else:
        click.echo(click.style("CEP not found", fg="blue"), err=True)
        sys.exit(3)



import logging
import shutil
import tempfile
import urllib.error
import urllib.request
import zipfile
from pathlib import Path
from typing import TYPE_CHECKING, Union
from urllib.parse import urlparse

from .exc import DneResolverError
from .table_set import TableSetEnum, get_table_files_glob

if TYPE_CHECKING:
    from zipfile import ZipFile

logger = logging.getLogger(__name__)

DELIMITED_SUBDIR = "Delimitado"


LATEST_DNE_DOWNLOAD_URL = (
    "https://www2.correios.com.br/sistemas/edne/download/eDNE_Basico.zip"
)


class DneResolver:
    """
    Resolves the DNE source path.

    The provided dne_source string can be:
        - None: Then the latest DNE will be downloaded from Correios website
        - A ZIP file URL: Then the DNE will be downloaded and extracted
        - A local ZIP file: Then the file will be extracted
        - A local folder: Then the folder will be validated and used as the DNE source

    Returns the path to the resolved DNE folder.
    """

    dne_source: Union[str, None]

    def __init__(self, dne_source: Union[str, None] = None):
        self.dne_source = dne_source

        # keep track of temporary files to be removed
        self.temp_artifacts = []

    def __enter__(self):
        try:
            logger.info("Resolving DNE source...", extra={"indentation": 0})
            return self.resolve_dne_source(self.dne_source)
        except Exception:
            self.remove_temp_artifacts()
            raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val and self.temp_artifacts:
            logger.warning(
                "Something went wrong. Removing temporary files...",
                extra={"indentation": 0},
            )
        self.remove_temp_artifacts()

    def resolve_dne_source(self, dne_source: str):
        if dne_source is None:
            logger.info(
                "No DNE source provided, the latest DNE will be downloaded from "
                "Correios website",
                extra={"indentation": 0},
            )
            dne_source = LATEST_DNE_DOWNLOAD_URL

        # if the provided source looks like a URL, download it
        if looks_like_a_url(dne_source):
            logger.debug("Source identified as URL")
            dne_source = self.download_dne(dne_source)

            logger.debug('DNE downloaded to "%s"', dne_source)

        # if the provided source is a file, try to unzip it
        if Path(dne_source).is_file():
            dne_source = self.resolve_file_source(dne_source)
            # remove no more needed temp files/dirs
            self.remove_temp_artifacts(keep_last=True)

        if (dne_dir := Path(dne_source)).is_dir():
            return self.resolve_dir_source(dne_dir)

        msg = f"DNE source not found: {dne_source}"
        raise DneResolverError(msg)

    def resolve_file_source(self, dne_source: str):
        try:
            with zipfile.ZipFile(dne_source, mode="r") as archive:
                logger.debug("Source identified as ZIP file: %s", dne_source)
                # At this point we know it's a ZIP file, but it can be one of:
                # 1. A ZIP file downloaded from Correios website containing two ZIP
                # files:
                #    - eDNE_Basico_YYMM.zip
                #    - eDNE_Delta_Basico_YYMM.zip
                # 2. The eDNE_Basico_YYMM.zip file extracted from the ZIP file above

                if dne_basico_filename := zip_contains_dne_basico_zip_file(archive):
                    logger.debug(
                        "Source is as ZIP file containing a DNE Basico ZIP file"
                    )
                    # if the ZIP file contains a DNE Basico ZIP file, extract it
                    temp_dir = tempfile.mkdtemp()
                    self.temp_artifacts.append(temp_dir)

                    logger.debug(
                        'Extracting "%s" to "%s"',
                        dne_basico_filename,
                        temp_dir,
                    )
                    extracted_zip = archive.extract(dne_basico_filename, temp_dir)

                    # start the resolver again with the extracted ZIP file as source
                    return self.resolve_dne_source(extracted_zip)

                else:
                    # the ZIP file isn't a ZIP file containing other ZIP files,
                    # so let's check if it's a DNE Basico ZIP file

                    valid_dne_files = [
                        f
                        for f in archive.namelist()
                        if filename_is_a_dne_basico_file(f)
                    ]

                    if valid_dne_files:
                        temp_dir = tempfile.mkdtemp()
                        self.temp_artifacts.append(temp_dir)

                        logger.debug(
                            "Source is a DNE Basico ZIP file, extracting %s "
                            "files to %s",
                            len(valid_dne_files),
                            temp_dir,
                        )

                        for file in valid_dne_files:
                            archive.extract(file, temp_dir)

                        return temp_dir
                    else:
                        msg = "ZIP file does not contain DNE Basico files"
                        raise DneResolverError(msg)

        except zipfile.BadZipFile as e:
            msg = f"Source is not a valid ZIP file: {dne_source}"
            raise DneResolverError(msg) from e

    def resolve_dir_source(self, dne_dir: Path):
        # assert all the data files are present
        for table in TableSetEnum.ALL_TABLES.to_populate:
            # check if there are source files for all tables to be created
            if (file_glob := get_table_files_glob(table)) and not any(
                dne_dir.glob(file_glob)
            ):
                if (delimited_subdir := (dne_dir / DELIMITED_SUBDIR)).is_dir():
                    return self.resolve_dne_source(str(delimited_subdir))

                msg = f"DNE data file not found: {dne_dir / file_glob}"
                raise DneResolverError(msg)

        return dne_dir

    def download_dne(self, url: str) -> str:
        """
        Download zipped DNE and returns the path to the downloaded file.
        """
        bs = 1024 * 8

        try:
            with urllib.request.urlopen(url) as response:  # noqa: S310
                total_size = int(response.headers.get("Content-Length", -1))
                self.download_report_hook(0, total_size, "start")

                with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                    # add downloaded file to the list of files to be removed
                    self.temp_artifacts.append(tmp_file.name)

                    while block := response.read(bs):
                        tmp_file.write(block)
                        self.download_report_hook(len(block), total_size, "progress")

                    self.download_report_hook(0, total_size, "finish")

        except urllib.error.URLError as e:
            msg = f"Failed to download DNE from {url}"
            raise DneResolverError(msg) from e

        return tmp_file.name

    def download_report_hook(self, read: int, total: int, hook_type: str):
        pass

    def remove_temp_artifacts(self, *, keep_last: bool = False):
        """
        Remove all temporary files created by the resolver.
        """

        while num_artifacts := len(self.temp_artifacts):
            if num_artifacts < 2 and keep_last:  # noqa: PLR2004
                return

            to_remove = self.temp_artifacts.pop(0)

            if Path(to_remove).is_file():
                logger.debug("Removing temporary file %s", to_remove)
                Path(to_remove).unlink()
            else:
                logger.debug("Removing temporary directory %s", to_remove)
                shutil.rmtree(to_remove)


def looks_like_a_url(url):
    """
    Rudimentar check to see if a string looks like a URL.
    """
    result = urlparse(url)
    return all([result.scheme, result.netloc, result.scheme in ("http", "https")])


def zip_contains_dne_basico_zip_file(zipfile: "ZipFile") -> Union[str, None]:
    """
    Check if the provided ZIP file contains a DNE Basico ZIP file.
    If true, returns the path to the DNE Basico ZIP file.
    """
    for filename in zipfile.namelist():
        lowered_name = filename.lower()
        if lowered_name.startswith("edne_basico_") and lowered_name.endswith(".zip"):
            return filename

    return None


def filename_is_a_dne_basico_file(filename: str) -> bool:
    """
    Check if the provided filename is part of a DNE Basico package.
    """
    name = filename.lower()
    return name == "leiame.txt" or (
        name.startswith(f"{DELIMITED_SUBDIR.lower()}/") and name.endswith(".txt")
    )


# def extract_dne_version(dne_path: Path) -> int:
#     """
#     Extracts the DNE version from the provided path
#     by reading it from the LEIAME.TXT file
#     """
#     leiame_file = dne_path / Path("LEIAME.TXT")
#
#     if not leiame_file.is_file():
#         msg = f"Failed to extract DNE version. File not found: {leiame_file}"
#         raise DneResolverError(msg)
#
#     content = leiame_file.open(encoding="latin1").read()
#     match = re.compile(r"DNE versão (?P<version>\d{5})").search(content)
#
#     if not match:
#         msg = f"Failed to extract DNE version. Version not found in {leiame_file}"
#         raise DneResolverError(msg)
#
#     return int(match.group("version"))



import logging
from typing import Iterable, List, Union

import sqlalchemy as sa

from .tables import metadata
from .unified_table import populate_unified_table

logger = logging.getLogger(__name__)


class DneDatabaseWriter:
    engine: sa.Engine
    connection: sa.Connection
    metadata: sa.MetaData = metadata
    insert_buffer_size = 1000

    def __init__(self, database_url: str):
        self.engine = sa.create_engine(database_url, echo=False)

    def __enter__(self):
        logger.info("Connecting to database...", extra={"indentation": 0})
        self.connection = self.engine.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            # if something went wrong, rollback all the changes
            self.connection.rollback()

        else:
            self.connection.commit()

        self.connection.close()

    def create_tables(self, tables: List[str]):
        metadata_tables = [self.metadata.tables[t] for t in tables]
        tables_names = "\n".join([f"- {t}" for t in tables])

        logger.info("Creating tables:\n%s", tables_names, extra={"indentation": 0})
        metadata.create_all(self.engine, tables=metadata_tables)

    def clean_tables(self, tables: List[str]):
        logger.info("Cleaning tables", extra={"indentation": 0})

        # delete rows in reverse order to avoid foreign key constraint violations
        for table_name in reversed(tables):
            table = self.metadata.tables[table_name]

            if num_rows := self.connection.execute(
                sa.select(sa.func.count()).select_from(table)
            ).scalar():
                logger.info(
                    "Deleting %s rows from table %s",
                    num_rows,
                    table.name,
                    extra={"indentation": 1},
                )
                self.connection.execute(table.delete())

    def drop_tables(self, tables: List[str]):
        if tables:
            logger.info("Dropping tables", extra={"indentation": 0})

            for table in reversed(tables):
                logger.info("Dropping table %s", table, extra={"indentation": 1})
                self.metadata.tables[table].drop(self.connection, checkfirst=True)

    def populate_table(self, table_name: str, lines: Iterable[List[str]]):
        logger.info("Populating table %s", table_name, extra={"indentation": 0})
        table = self.metadata.tables[table_name]
        columns = [c.name for c in table.columns]

        self_referencing_fk = self.find_self_referencing_fks(table)

        if self_referencing_fk:
            # if the table has a self-referencing foreign key, the rows
            # need to be sorted in a way the ancestors are inserted first
            lines = self.sort_topologically(lines, self_referencing_fk, columns)

        buffer = []

        for count, line in enumerate(lines, start=1):  # noqa: B007
            buffer.append(dict(zip(columns, line)))

            if len(buffer) >= self.insert_buffer_size:
                self.connection.execute(table.insert(), buffer)
                buffer = []

        if buffer:
            self.connection.execute(table.insert(), buffer)

        logger.info(
            'Inserted %s rows into table "%s"',
            count,
            table_name,
            extra={"indentation": 1},
        )

    def populate_unified_table(self):
        logger.info("Populating unified CEP table", extra={"indentation": 0})
        populate_unified_table(self.connection)

    @staticmethod
    def find_self_referencing_fks(table) -> Union[str, None]:
        """
        Find any column that references the same table.
        If there is one, the rows must be topologically sorted before inserting.
        """
        for fk in list(table.foreign_keys):
            if fk.column.table.name == table.key:
                return fk.parent.name
        return None

    @staticmethod
    def sort_topologically(
        lines: Iterable[List[str]], self_referencing_fk: str, columns: List[str]
    ) -> Iterable[List[str]]:
        from graphlib import TopologicalSorter

        fk_index = columns.index(self_referencing_fk)

        # convert iterable to list as it will be iterated multiple times
        lines = list(lines)

        topological_graph = {}

        for line in lines:
            topological_graph.setdefault(line[0], set()).add(line[fk_index])

        sorted_map = tuple(TopologicalSorter(topological_graph).static_order())
        return sorted(lines, key=lambda line: sorted_map.index(line[0]))



from typing import Union

from sqlalchemy import create_engine

from .tables import metadata


class CepQuerier:
    cep_table = metadata.tables["cep_unificado"]

    def __init__(self, database_url: str):
        self.engine = create_engine(database_url, echo=False)

    def query(self, cep: str) -> Union[dict, None]:
        cep = cep.replace("-", "").strip()

        with self.engine.connect() as conn:
            cep = conn.execute(
                self.cep_table.select().where(self.cep_table.c.cep == cep)
            ).first()

            return cep._asdict() if cep else None



import enum

from sqlalchemy import (
    Column,
    Enum,
    ForeignKey,
    Integer,
    MetaData,
    String,
    Table,
)

metadata = MetaData()


"""
Faixa de CEP de UF
"""
log_faixa_uf = Table(
    "log_faixa_uf",
    metadata,
    Column(
        "ufe_sg",
        String(2),
        primary_key=True,
        comment="Sigla da UF",
        nullable=False,
    ),
    Column(
        "ufe_cep_ini",
        String(8),
        primary_key=True,
        comment="CEP inicial da UF",
        nullable=False,
    ),
    Column("ufe_cep_fim", String(8), comment="CEP final da UF", nullable=False),
)


class SituacaoLocalidadeEnum(enum.Enum):
    """
    Situação da localidade

    0 - Localidade não codificada ao nível de Logradouro
    1 - Localidade codificada ao nível de Logradouro
    2 - Distrito ou Povoado inserido na codificação ao nível de Logradouro
    3 - Localidade em fase de codificação ao nível de Logradouro
    """

    def __str__(self):
        return self.value

    NAO_CODIFICADA = "0"
    CODIFICADA = "1"
    DISTRITO = "2"
    FASE_CODIFICACAO = "3"


class TipoLocalidadeEnum(enum.Enum):
    """
    Tipo de localidade
    """

    def __str__(self):
        return self.value

    DISTRITO = "D"
    MUNICIPIO = "M"
    POVOADO = "P"


"""
Localidade

O arquivo LOG_LOCALIDADE contempla os municípios, distritos e povoados do Brasil.

Os CEPs presentes neste arquivo valem para todos os logradouros da cidade, não
necessitando consulta nos demais arquivos.

As localidades em fase de codificação(LOC_IN_SIT=3) estão em período de transição,
sendo aceito o CEP Geral ou os CEPs de Logradouros para endereçamento.
"""
log_localidade = Table(
    "log_localidade",
    metadata,
    Column(
        "loc_nu",
        Integer,
        primary_key=True,
        unique=True,
        comment="Chave da localidade",
        nullable=False,
    ),
    Column(
        "ufe_sg",
        String(2),
        primary_key=True,
        comment="Sigla da UF",
        nullable=False,
    ),
    Column("loc_no", String(72), comment="Nome da localidade", nullable=False),
    # CEP da localidade(para localidade não codificada, ou seja, loc_in_sit = 0)
    Column("cep", String(8), index=True, comment="CEP da localidade"),
    Column(
        "loc_in_sit",
        Enum(
            SituacaoLocalidadeEnum,
            values_callable=lambda x: [e.value for e in x],
        ),
        comment="Situação da localidade",
        nullable=False,
    ),
    Column(
        "loc_in_tipo_loc",
        Enum(TipoLocalidadeEnum, values_callable=lambda x: [e.value for e in x]),
        comment="Tipo de localidade",
        nullable=False,
    ),
    Column(
        "loc_nu_sub",
        ForeignKey("log_localidade.loc_nu", ondelete="CASCADE"),
        index=True,
        comment="Chave da localidade de subordinação",
    ),
    Column("loc_no_abrev", String(36), comment="Abreviatura do nome da localidade"),
    Column("mun_nu", Integer, comment="Código do município IBGE"),
)


"""
Outras denominações da Localidade (denominação popular, denominação anterior)
"""
log_var_loc = Table(
    "log_var_loc",
    metadata,
    Column(
        "loc_nu",
        ForeignKey(log_localidade.c.loc_nu),
        primary_key=True,
        comment="Chave da localidade",
    ),
    Column(
        "val_nu",
        Integer,
        primary_key=True,
        comment="Ordem da denominação",
    ),
    Column("val_tx", String(72), comment="Denominação", nullable=False),
)


class TipoFaixaCepEnum(enum.Enum):
    """
    Tipo de faixa de CEP

    T - Total do município
    C - Exclusiva da sede urbana
    """

    TOTAL_MUNICIPIO = "T"
    EXCLUSIVA_SEDE_URBANA = "C"


"""
Faixa de CEP das Localidades codificadas

Este arquivo contém dados relativos às faixas de CEP das localidades classificadas na
categoria político-administrativa de município codificadas com CEP único ou codificadas
por logradouros.
"""
log_faixa_localidade = Table(
    "log_faixa_localidade",
    metadata,
    Column(
        "loc_nu",
        ForeignKey(log_localidade.c.loc_nu),
        primary_key=True,
        comment="Chave da localidade",
    ),
    Column(
        "loc_cep_ini",
        String(8),
        primary_key=True,
        comment="CEP inicial da localidade",
    ),
    Column("loc_cep_fim", String(8), comment="CEP final da localidade"),
    Column(
        "loc_tipo_faixa",
        Enum(TipoFaixaCepEnum, values_callable=lambda x: [e.value for e in x]),
        primary_key=True,
        comment="Tipo de faixa de CEP",
    ),
)

"""
Bairro
"""
log_bairro = Table(
    "log_bairro",
    metadata,
    Column("bai_nu", Integer, primary_key=True, comment="Chave do bairro"),
    Column("ufe_sg", String(2), comment="Sigla da UF", nullable=False),
    Column(
        "loc_nu",
        ForeignKey(log_localidade.c.loc_nu),
        index=True,
        comment="Chave da localidade",
        nullable=False,
    ),
    Column("bai_no", String(72), comment="Nome do bairro", nullable=False),
    Column(
        "bai_no_abrev",
        String(36),
        comment="Abreviatura do nome do bairro",
    ),
)

"""
Outras denominações do Bairro Localidade (denominação popular, denominação anterior)
"""
log_var_bai = Table(
    "log_var_bai",
    metadata,
    Column(
        "bai_nu",
        ForeignKey(log_bairro.c.bai_nu),
        primary_key=True,
        comment="Chave do bairro",
    ),
    Column("vdb_nu", Integer, primary_key=True, comment="Ordem da denominação"),
    Column("vdb_tx", String(72), comment="Denominação", nullable=False),
)


"""
Faixa de CEP de Bairro
"""
log_faixa_bairro = Table(
    "log_faixa_bairro",
    metadata,
    Column(
        "bai_nu",
        ForeignKey(log_bairro.c.bai_nu),
        primary_key=True,
        comment="Chave do bairro",
    ),
    Column(
        "fcb_cep_ini",
        String(8),
        primary_key=True,
        comment="CEP inicial do bairro",
    ),
    Column("fcb_cep_fim", String(8), comment="CEP final do bairro", nullable=False),
)


"""
Caixa Postal Comunitária (CPC)

São áreas rurais e/ou urbanas periféricas não atendidas pela distribuição domiciliária.
"""
log_cpc = Table(
    "log_cpc",
    metadata,
    Column(
        "cpc_nu",
        Integer,
        primary_key=True,
        comment="Chave da caixa postal comunitária",
    ),
    Column("ufe_sg", String(2), comment="Sigla da UF", nullable=False),
    Column(
        "loc_nu",
        ForeignKey(log_localidade.c.loc_nu),
        index=True,
        comment="Chave da localidade",
        nullable=False,
    ),
    Column("cpc_no", String(72), comment="Nome da CPC", nullable=False),
    Column("cpc_endereco", String(100), comment="Endereço da CPC", nullable=False),
    Column("cep", String(8), index=True, comment="CEP da CPC", nullable=False),
)

"""
Faixa de Caixa Postal Comunitária
"""
log_faixa_cpc = Table(
    "log_faixa_cpc",
    metadata,
    Column(
        "cpc_nu",
        ForeignKey(log_cpc.c.cpc_nu),
        primary_key=True,
        comment="Chave da caixa postal comunitária",
    ),
    Column(
        "cpc_inicial",
        String(6),
        comment="Número inicial da caixa postal comunitária",
        primary_key=True,
    ),
    Column(
        "cpc_final",
        String(6),
        comment="Número final da caixa postal comunitária",
        primary_key=True,
    ),
)


"""
Logradouro

Este arquivo contém os registros das localidades codificadas por logradouro
(LOC_IN_SIT=1) e de localidades em fase de codificação (LOC_IN_SIT=3).

Para encontrar o bairro do logradouro, utilize o campo BAI_NU_INI
(relacionamento com LOG_BAIRRO, campo BAI_NU)
"""
log_logradouro = Table(
    "log_logradouro",
    metadata,
    Column("log_nu", Integer, primary_key=True, comment="Chave do logradouro"),
    Column("ufe_sg", String(2), comment="Sigla da UF", nullable=False),
    Column(
        "loc_nu",
        ForeignKey(log_localidade.c.loc_nu),
        index=True,
        comment="Chave da localidade",
        nullable=False,
    ),
    Column(
        "bai_nu_ini",
        ForeignKey(log_bairro.c.bai_nu),
        index=True,
        comment="Chave do bairro inicial do logradouro",
        nullable=False,
    ),
    Column(
        "bai_nu_fim",
        ForeignKey(log_bairro.c.bai_nu),
        index=True,
        comment="Chave do bairro final do logradouro",
        nullable=True,
    ),
    Column("log_no", String(100), comment="Nome do logradouro", nullable=False),
    Column("log_complemento", String(100), comment="Complemento"),
    Column(
        "cep",
        String(8),
        index=True,
        comment="CEP do logradouro",
        nullable=False,
    ),
    Column("tlo_tx", String(36), comment="Tipo de logradouro", nullable=False),
    Column(
        "log_sta_tlo",
        String(1),
        index=True,
        comment="Indicador de utilização do tipo de logradouro (S/N)",
    ),
    Column("log_no_abrev", String(36), comment="Abreviatura do nome do logradouro"),
    info={"file_glob": "LOG_LOGRADOURO_*.TXT"},
)


"""
Outras denominações do Logradouro (denominação popular, denominação anterior)
"""
log_var_log = Table(
    "log_var_log",
    metadata,
    Column(
        "log_nu",
        ForeignKey(log_logradouro.c.log_nu),
        primary_key=True,
        comment="Chave do logradouro",
    ),
    Column("vlo_nu", Integer, primary_key=True, comment="Ordem da denominação"),
    Column("tlo_tx", String(36), comment="Tipo de logradouro da variação"),
    Column("vlo_tx", String(150), comment="Nome da variação do logradouro"),
)


class SeccionamentoLadoEnum(enum.Enum):
    """
    Indica a paridade/lado do seccionamento
    """

    AMBOS = "A"
    PAR = "P"
    IMPAR = "I"
    DIREITO = "D"
    ESQUERDO = "E"


"""
Faixa numérica do seccionamento
"""
log_num_sec = Table(
    "log_num_sec",
    metadata,
    Column(
        "log_nu",
        ForeignKey(log_logradouro.c.log_nu),
        primary_key=True,
        comment="Chave do logradouro",
    ),
    Column("sec_nu_ini", String(10), comment="Número inicial do seccionamento"),
    Column("sec_nu_fim", String(10), comment="Número final do seccionamento"),
    Column(
        "sec_in_lado",
        Enum(
            SeccionamentoLadoEnum,
            values_callable=lambda x: [e.value for e in x],
        ),
        comment="Indica a paridade/lado do seccionamento",
    ),
)


"""
Grande usuário

São clientes com grande volume postal (empresas, universidades, bancos,
órgãos públicos, etc).

O campo LOG_NU está sem conteúdo para as localidades não codificadas(LOC_IN_SIT=0),
devendo ser utilizado o campo GRU_ENDEREÇO para  endereçamento.
"""
log_grande_usuario = Table(
    "log_grande_usuario",
    metadata,
    Column("gru_nu", Integer, primary_key=True, comment="Chave do grande usuário"),
    Column("ufe_sg", String(2), comment="Sigla da UF", nullable=False),
    Column(
        "loc_nu",
        ForeignKey(log_localidade.c.loc_nu),
        index=True,
        comment="Chave da localidade",
        nullable=False,
    ),
    Column(
        "bai_nu",
        ForeignKey(log_bairro.c.bai_nu),
        index=True,
        comment="Chave do bairro",
        nullable=False,
    ),
    Column(
        "log_nu",
        ForeignKey(log_logradouro.c.log_nu),
        index=True,
        comment="Chave do logradouro",
    ),
    Column("gru_no", String(72), comment="Nome do grande usuário", nullable=False),
    Column(
        "gru_endereco",
        String(100),
        comment="Endereço do grande usuário",
        nullable=False,
    ),
    Column(
        "cep",
        String(8),
        index=True,
        comment="CEP do grande usuário",
        nullable=False,
    ),
    Column(
        "gru_no_abrev",
        String(36),
        comment="Abreviatura do nome do grande usuário",
    ),
)

"""
Unidade Operacional dos Correios.

São agências próprias ou terceirizadas, centros de distribuição, etc.
O campo LOG_NU está sem conteúdo para as localidades não codificadas(LOC_IN_SIT=0),
devendo ser utilizado o campo UOP_ENDEREÇO para endereçamento.
"""
log_unid_oper = Table(
    "log_unid_oper",
    metadata,
    Column(
        "uop_nu",
        Integer,
        primary_key=True,
        comment="Chave da unidade operacional",
    ),
    Column("ufe_sg", String(2), comment="Sigla da UF", nullable=False),
    Column(
        "loc_nu",
        ForeignKey(log_localidade.c.loc_nu),
        index=True,
        comment="Chave da localidade",
        nullable=False,
    ),
    Column(
        "bai_nu",
        ForeignKey(log_bairro.c.bai_nu),
        index=True,
        comment="Chave do bairro",
        nullable=False,
    ),
    Column(
        "log_nu",
        ForeignKey(log_logradouro.c.log_nu),
        index=True,
        comment="Chave do logradouro",
    ),
    Column(
        "uop_no",
        String(100),
        comment="Nome da unidade operacional",
        nullable=False,
    ),
    Column(
        "uop_endereco",
        String(100),
        comment="Endereço da unidade operacional",
        nullable=False,
    ),
    Column(
        "cep",
        String(8),
        index=True,
        comment="CEP da unidade operacional",
        nullable=False,
    ),
    Column(
        "uop_in_cp",
        String(1),
        comment="Indicador de unidade operacional com caixa postal (S/N)",
    ),
    Column(
        "uop_no_abrev",
        String(36),
        comment="Abreviatura do nome da unidade operacional",
    ),
)


"""
Faixa de Caixa Postal - UOP
"""
log_faixa_uop = Table(
    "log_faixa_uop",
    metadata,
    Column(
        "upo_nu",
        ForeignKey(log_unid_oper.c.uop_nu),
        primary_key=True,
        comment="Chave da unidade operacional",
    ),
    Column(
        "fnc_inicial",
        Integer,
        primary_key=True,
        comment="Número inicial da caixa postal",
    ),
    Column(
        "fnc_final",
        Integer,
        comment="Número final da caixa postal",
        nullable=False,
    ),
)


"""
Relação dos Nomes dos Países, suas siglas e grafias em inglês e francês
"""
ect_pais = Table(
    "ect_pais",
    metadata,
    Column("pai_sg", String(2), primary_key=True, comment="Sigla do país"),
    Column("pai_sg_alternativa", String(3), comment="Sigla alternativa do país"),
    Column("pai_no_portugues", String(72), comment="Nome do país em português"),
    Column("pai_no_ingles", String(72), comment="Nome do país em inglês"),
    Column("pai_no_frances", String(72), comment="Nome do país em francês"),
    Column("pai_abreviatura", String(36), comment="Abreviatura do nome do país"),
)


"""
Tabela unificada de CEP

Não inclusa no DNE, é populada com dados das outras tabelas depois que a importação
é concluída.
"""
cep_unificado = Table(
    "cep_unificado",
    metadata,
    Column("cep", String(8), primary_key=True),
    Column("logradouro", String(100)),
    Column("complemento", String(100)),
    Column("bairro", String(72)),
    Column("municipio", String(72), nullable=False),
    Column("municipio_cod_ibge", Integer, nullable=False),
    Column("uf", String(2), nullable=False),
    # In some special cases, a CEP is assigned to a single address and may have a name.
    # That happens for government agencies, large clients, some condos,
    # Correios' own buildings, etc.
    Column("nome", String(100)),
    info={"unified_table": True},
)



import logging
from pathlib import Path
from typing import Iterable, Type, Union

from .dbwriter import DneDatabaseWriter
from .resolver import DneResolver
from .table_set import TableSetEnum, get_table_files_glob

logger = logging.getLogger(__name__)


class DneLoader:
    DneResolver: Type[DneResolver] = DneResolver
    DneDatabaseWriter: Type[DneDatabaseWriter] = DneDatabaseWriter

    database_url: str
    dne_source: str

    read_buffer_size = 1000000

    def __init__(
        self,
        database_url: str,
        *,
        dne_source: Union[str, None] = None,
    ):
        self.database_url = database_url
        self.dne_source = dne_source

    def load(self, table_set: TableSetEnum = TableSetEnum.UNIFIED_CEP_ONLY):
        # connect to database to ensure the URL is valid
        # connection will be closed when the context manager exits
        with self.DneDatabaseWriter(self.database_url) as database_writer:
            # now that we know the URL is valid, download/extract the DNE file
            # temp files will be removed when the context manager exits
            with self.DneResolver(self.dne_source) as dne_path:
                # all good, let's start by ensuring the tables exist and are empty
                database_writer.create_tables(table_set.to_populate)
                database_writer.clean_tables(table_set.to_populate)

                for table in table_set.to_populate:
                    files_glob = get_table_files_glob(table)

                    if files_glob:
                        files = dne_path.glob(files_glob)
                        data = TableFilesReader(
                            files, buffer_size=self.read_buffer_size
                        )
                        database_writer.populate_table(table, data)

            database_writer.populate_unified_table()
            database_writer.drop_tables(table_set.to_drop)


class TableFilesReader:
    """
    Memory-efficient reader for DNE files targeting a single table.
    Read files sequentially in chunks of lines and yield each line
    """

    def __init__(self, files: Iterable[Path], buffer_size=1000000):
        self.files = files
        self.buffer_size = buffer_size

    def __iter__(self):
        for file in self.files:
            with file.open(encoding="latin1") as fp:
                logger.info("Reading %s", file.name, extra={"indentation": 1})
                lines_buffer = fp.readlines(self.buffer_size)

                while lines_buffer:
                    logger.debug(
                        "Read %s lines from %s",
                        len(lines_buffer),
                        file.name,
                        extra={"indentation": 2},
                    )
                    for line in lines_buffer:
                        yield [f.strip() or None for f in line.split("@")]

                    lines_buffer = fp.readlines(self.buffer_size)
